<?php

namespace Drupal\elasticsearch_connector_suggester\EventSubscriber;

use Drupal\Core\Config\ConfigEvents;
use Drupal\Core\Entity\EntityInterface;
use Drupal\elasticsearch_connector\ElasticSearch\Parameters\Factory\IndexFactory;
use Drupal\Component\EventDispatcher\Event;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\elasticsearch_connector\Event\BuildIndexParamsEvent;
use Drupal\elasticsearch_connector\Event\BuildSearchParamsEvent;
use Drupal\elasticsearch_connector\Event\PrepareIndexMappingEvent;
use Drupal\search_api\Event\SearchApiEvents;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Class DefaultSubscriber for elasticsearch connector suggester.
 */
class DefaultSubscriber implements EventSubscriberInterface {

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\Entity
   */
  private $entityTypeManager;

  /**
   * Constructs a new DefaultSubscriber object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_typeManager
   *   Entity type manager interface.
   */
  public function __construct(EntityTypeManagerInterface $entity_typeManager) {
    $this->entityTypeManager = $entity_typeManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[PrepareIndexMappingEvent::PREPARE_INDEX_MAPPING] = [
      'onElasticsearchConnectorPrepareIndexMapping',
    ];
    $events[BuildIndexParamsEvent::BUILD_PARAMS] = [
      'onElasticsearchBuildParams',
    ];
    $events[BuildSearchParamsEvent::BUILD_QUERY] = [
      'onElasticsearchBuildQuery',
    ];
    $events[ConfigEvents::SAVE] = [
      'onConfigSave',
    ];
    $events[SearchApiEvents::PROCESSING_RESULTS] = [
      'onSearchApiProcessingResults',
    ];

    return $events;
  }

  /**
   * Called on elasticsearch_connector.prepare_index_mapping event.
   *
   * Add suggester field to index mapping.
   *
   * @param \Drupal\elasticsearch_connector\Event\PrepareIndexMappingEvent $event
   *   The event.
   */
  public function onElasticsearchConnectorPrepareIndexMapping(Event $event) {
    $fields = $this->getFieldsByEvent($event);
    if (empty($fields)) {
      return;
    }
    $params = $event->getIndexMappingParams();
    foreach ($fields as $field_id) {
      // Set suggester field.
      $params['body']['properties'][$field_id . '_suggest'] = [
        'type' => 'completion',
        'analyzer' => 'standard',
        'search_analyzer' => 'standard',
        'preserve_separators' => TRUE,
        'preserve_position_increments' => TRUE,
        'max_input_length' => 1000,
      ];
    }

    $event->setIndexMappingParams($params);
  }

  /**
   * Called on elasticsearch_connector.build_params event.
   *
   * Add suggester value to index element.
   *
   * @param \Drupal\elasticsearch_connector\Event\BuildIndexParamsEvent $event
   *   The event.
   */
  public function onElasticsearchBuildParams(Event $event) {
    $fields = $this->getFieldsByEvent($event);
    if (empty($fields)) {
      return;
    }


    $params = $event->getElasticIndexParams();
    foreach ($fields as $field) {
      $suggested_name = $field . '_suggest';
      foreach ($params['body'] as &$body) {
        if (!isset($body[$field])) {
          continue;
        }
        $text = $body[$field] ?? '';
        if (is_array($text)) {
          $text = current($text);
        }
        // Set suggester value.
        $body[$suggested_name][] = [
          'input' => $text,
          'weight' => 1,
        ];
        $parts = explode(' ', $text);
        $count = count($parts);
        if ($count > 1) {
          // Set suggester value for each word.
          for ($i = 1; $i < $count; $i++) {
            $part = array_slice($parts, 0, $i);
            if (count($part) > 1) {
              $body[$suggested_name][] = [
                'input' => implode(' ', $part),
                'weight' => 3,
              ];
            }
            $word = array_pop($part);
            $body[$suggested_name][] = [
              'input' => $word,
              'weight' => 5,
            ];
          }
        }
      }
    }

    $event->setElasticIndexParams($params);
  }

  /**
   * Called on elasticsearch_connector.prepare_query event.
   *
   * Replace query to suggester request.
   *
   * @param \Drupal\elasticsearch_connector\Event\BuildSearchParamsEvent $event
   *   The event.
   */
  public function onElasticsearchBuildQuery(Event $event) {
    $params = $event->getElasticSearchParams();
    $index_name = $params["index"] ?? NULL;
    if ($index_name === NULL) {
      return;
    }
    $fields = $this->getFields($index_name);
    if (empty($fields)) {
      return;
    }
    $field_name = $params['body']['query']['bool']['must']['query_string']['fields'][0] ?? '';
    if (empty($field_name)) {
      return;
    }
    $query_string = $params['body']['query']['bool']['must']['query_string']['query'] ?? '';
    if (empty($query_string)) {
      return;
    }
    $string = str_replace('AND ', '', $query_string);
    if (empty($string)) {
      return;
    }

    // Prepare suggester field name.
    $name = $field_name;
    if (strpos($field_name, '_suggest') === FALSE) {
      // Remove from name the ^0 or same string replace with _suggest string.
      $name = preg_replace('/\^.*$/', '_suggest', $field_name);
    }
    if ($name === $field_name && strpos($name, '_suggest') === FALSE) {
      // Field name not contains _suggest string.
      $name .= '_suggest';
    }

    // Set suggest request.
    $params['body'] = [
      '_source' => 'suggest',
      'suggest' => [
        'autocomplete' => [
          'prefix' => $string,
          'completion' => [
            'field' => $name,
            'size' => 10,
          ],
        ],
      ],
    ];

    $event->setElasticSearchParams($params);
  }

  /**
   * Called on config.save event.
   *
   * Update elasticsearch index mapping.
   *
   * @param \Drupal\Core\Config\ConfigEvents $event
   *   The event.
   */
  public function onConfigSave(Event $event) {
    /** @var \Drupal\Core\Config\Config $config */
    $config = $event->getConfig();
    $name = $config->getName();
    if (!preg_match('/^search_api_autocomplete.search.*$/', $name)) {
      return;
    }
    // Save index to update mapping.
    if ($index = $this->entityTypeManager->getStorage('search_api_index')
      ->load($config->get('index_id'))) {
      $index->save();
    }
  }

  /**
   * Called on search_api.processing_results event.
   *
   * Add suggester results.
   *
   * @param \Drupal\search_api\Event\ProcessingResultsEvent $event
   *   The event.
   */
  public function onSearchApiProcessingResults(Event $event) {
    /** @var \Drupal\search_api\Query\ResultSet $results */
    $results = $event->getResults();
    if (!empty($results->getResultItems())) {
      return;
    }
    $query = $results->getQuery();
    $index = $query->getIndex();
    /** @var \Drupal\search_api\Utility\FieldsHelper $fields_helper */
    $fields_helper = \Drupal::getContainer()->get('search_api.fields_helper');
    $response = $results->getExtraData('elasticsearch_response');
    $suggestions = $response['suggest']['autocomplete'][0]['options'] ?? [];
    array_map(function ($item) use ($fields_helper, $index, &$results) {
      $result_item = $fields_helper->createItem($index, $item['_id']);
      $result_item->setScore($item['_score']);
      $results->addResultItem($result_item);
    }, $suggestions);
  }

  /**
   * Get autocomplete fields.
   *
   * @param \Drupal\Component\EventDispatcher\Event $event
   *   The event.
   *
   * @return array
   *   The fields.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getFieldsByEvent(Event $event): array {
    $index_name = $event->getIndexName();
    return $this->getFields($index_name);
  }

  /**
   * Get autocomplete fields.
   *
   * @param string $index_name
   *   The index name.
   *
   * @return array
   *   The fields.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getFields(string $index_name): array {
    $fields = [];

    // Load all autocomplete entities.
    $autocompletes = $this->entityTypeManager
      ->getStorage('search_api_autocomplete_search')
      ->loadMultiple();
    $index = $this->loadIndexFromIndexName($index_name);

    /** @var \Drupal\search_api_autocomplete\Entity\Search $value */
    foreach ($autocompletes as $value) {
      // Get autocomplete settings for the index.
      if ($value->getIndexId() == $index->id()) {
        /** @var \Drupal\search_api_autocomplete\Entity\Search $autocomplete */
        $autocomplete = $value;
        break;
      }
    }

    if (!isset($autocomplete)) {
      // No autocomplete settings for the index.
      return $fields;
    }

    /** @var \Drupal\search_api_autocomplete\Suggester\SuggesterInterface $suggester */
    foreach ($autocomplete->getSuggesters() as $suggester) {
      if ($suggester->getPluginId() != 'elastic_live_results') {
        // Not elasticsearch suggester.
        continue;
      }
      // Get fields for the suggester.
      $fields[] = $suggester->getConfiguration()['fields'] ?? [];
    }

    return array_merge(...$fields);
  }

  /**
   * Loads the index entity associated with this event.
   *
   * @param string $index_name
   *   The long index name as a string.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   The loaded index or NULL.
   */
  private function loadIndexFromIndexName($index_name):? EntityInterface {
    /** @var \Drupal\search_api\Entity\Index[] $search_api_indexes */
    $search_api_indexes = $this->entityTypeManager
      ->getStorage('search_api_index')
      ->loadMultiple();

    foreach ($search_api_indexes as $search_api_index) {
      $elasticsearch_connector_index_name = IndexFactory::getIndexName($search_api_index);
      if ($index_name == $elasticsearch_connector_index_name) {
        return $search_api_index;
      }
    }

    return NULL;
  }

}
